/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Index - IndexingController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.controller.index;

import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.linkTo;
import static org.springframework.hateoas.server.mvc.WebMvcLinkBuilder.methodOn;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.hateoas.RepresentationModel;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpStatus;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestParam;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;

import ch.unige.solidify.business.IndexFieldAliasService;
import ch.unige.solidify.controller.IndexResourceController;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.index.indexing.IndexingService;
import ch.unige.solidify.model.SearchRequest;
import ch.unige.solidify.model.index.IndexFieldAlias;
import ch.unige.solidify.model.index.IndexMetadata;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.FacetPage;
import ch.unige.solidify.rest.FacetRequest;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.SearchCondition;
import ch.unige.solidify.rest.SearchConditionType;
import ch.unige.solidify.security.TrustedUserPermissions;
import ch.unige.solidify.util.StringTool;

public abstract class IndexingController<T extends IndexMetadata> extends IndexResourceController<String, T> {

  private static final Logger log = LoggerFactory.getLogger(IndexingController.class);

  private IndexFieldAliasService indexFieldAliasService;

  protected IndexingController(
          IndexingService<T> indexingService,
          IndexFieldAliasService indexFieldAliasService) {
    super(indexingService);
    this.indexFieldAliasService = indexFieldAliasService;
  }

  // **********************************
  // Override methods to apply security
  // **********************************
  @Override
  @TrustedUserPermissions
  public HttpEntity<T> create(@RequestBody T t) {
    return super.create(t);
  }

  @Override
  @TrustedUserPermissions
  public HttpEntity<T> update(@PathVariable String id, @RequestBody T t2) {
    return super.update(id, t2);
  }

  @Override
  @TrustedUserPermissions
  public HttpStatus delete(@PathVariable String id) {
    return super.delete(id);
  }

  @Override
  @TrustedUserPermissions
  public HttpStatus deleteList(@RequestBody String[] ids) {
    return super.deleteList(ids);
  }
  // **********************************

  @GetMapping("/" + ActionName.SEARCH)
  public HttpEntity<RestCollection<T>> search(
          @RequestParam(defaultValue = "*") String query,
          @RequestParam(required = false) String field,
          @RequestParam(required = false) String from,
          @RequestParam(required = false) String to,
          @RequestParam(required = false) String set,
          @RequestParam(required = false) String orgUnit,
          @RequestParam(required = false) String conditions,
          Pageable pageable) {

    /**
     * Build search conditions
     */
    List<SearchCondition> searchConditions = this.initSearchConditions(conditions);
    this.addSpecificCondition(searchConditions);
    this.addRangeCondition(searchConditions, field, from, to);
    this.addQueryCondition(searchConditions, query);

    // TODO : support one one criterion
    if (set != null && set.contains(":")) {
      String fieldQ = set.substring(0, set.lastIndexOf(':'));
      final String valueQ = set.substring(set.lastIndexOf(':'));
      fieldQ = fieldQ.replace(":", "\\:");
      set = fieldQ + valueQ;
    }
    this.addQueryCondition(searchConditions, set);

    /**
     * Get facets to return
     */
    List<FacetRequest> facetsList = this.indexFieldAliasService.getFacetsListForIndex(this.getIndex());

    /**
     * Do the search
     */
    final FacetPage<T> listItem = this.indexResourceService.search(this.getIndex(), searchConditions, facetsList, pageable);

    return this.page2Collection(listItem, pageable);
  }

  private List<SearchCondition> initSearchConditions(String conditions) {
    List<SearchCondition> searchConditions = new ArrayList<>();
    if (!StringTool.isNullOrEmpty(conditions)) {
      ObjectMapper mapper = new ObjectMapper();
      try {
        searchConditions = mapper.readValue(conditions, new TypeReference<List<SearchCondition>>() {
        });
        this.replaceAliasesByFieldNames(searchConditions);

      } catch (JsonProcessingException e) {
        throw new SolidifyRuntimeException("unable to parse search conditions", e);
      }
    }
    return searchConditions;
  }

  /**
   * Add specific search conditions
   *
   * @param searchConditions
   */
  protected abstract void addSpecificCondition(List<SearchCondition> searchConditions);

  /**
   * Look in database if there is field aliases corresponding to conditions -> transform condition from alias to real field name
   *
   * @param conditions
   */
  private void replaceAliasesByFieldNames(List<SearchCondition> conditions) {
    for (SearchCondition condition : conditions) {
      IndexFieldAlias fieldAlias = this.indexFieldAliasService.findByIndexNameAndAlias(this.getIndex(), condition.getField());
      if (fieldAlias != null) {
        condition.setField(fieldAlias.getField());
      }
    }
  }

  private void addRangeCondition(List<SearchCondition> searchConditions, String field, String from, String to) {
    if (!StringTool.isNullOrEmpty(field) && (!StringTool.isNullOrEmpty(from) || !StringTool.isNullOrEmpty(to))) {
      SearchCondition rangeCondition = new SearchCondition(SearchConditionType.RANGE, field);
      if (!StringTool.isNullOrEmpty(from)) {
        rangeCondition.setLowerValue(from);
      }
      if (!StringTool.isNullOrEmpty(to)) {
        rangeCondition.setUpperValue(to);
      }
      searchConditions.add(rangeCondition);
    }
  }

  private void addQueryCondition(List<SearchCondition> searchConditions, String query) {
    if (!StringTool.isNullOrEmpty(query)) {
      SearchCondition condition = new SearchCondition();
      condition.setType(SearchConditionType.QUERY);
      condition.setValue(query);
      searchConditions.add(condition);
    }
  }

  @PostMapping("/" + ActionName.SEARCH)
  public HttpEntity<RestCollection<T>> searchPost(@RequestBody SearchRequest searchRequest) {

    log.trace("Search parameters: {}", searchRequest);

    // Pageable
    final Pageable pageable;
    if (searchRequest.getSort() != null) {
      pageable = PageRequest.of(searchRequest.getPage(), searchRequest.getSize(), searchRequest.getSort());
    } else {
      pageable = PageRequest.of(searchRequest.getPage(), searchRequest.getSize());
    }

    // Search
    return this.search(searchRequest.getQuery(), null, null, null, null, null, searchRequest.getConditions(), pageable);
  }

  @Override
  protected <W extends RepresentationModel<W>> void addOthersLinks(W w) {
    w.add(linkTo(methodOn(this.getClass()).search(null, null, null, null, null, null, null, null)).withRel(ActionName.SEARCH));
  }

}
